# 
#   Function
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from base64 import b64encode
from filetype import guess_mime
from io import BytesIO
from pathlib import Path
from requests import put
from rich.progress import open as open_progress, wrap_file
from typing import Union

from ..graph import GraphClient
from ..types import UploadType

class StorageService:

    def __init__ (self, client: GraphClient) -> None:
        self.client = client

    def create_upload_url (self, name: str, type: UploadType, key: str=None) -> str:
        """
        Create an upload URL.

        Parameters:
            name (str): File name.
            type (UploadType): Upload type.
            key (str): File key. This is useful for grouping related files.

        Returns:
            str: File upload URL.
        """
        # Query
        response = self.client.query(f"""
            mutation ($input: CreateUploadUrlInput!) {{
                createUploadUrl (input: $input)
            }}
            """,
            { "input": { "type": type, "name": name, "key": key } }
        )
        # Return
        return response["createUploadUrl"]
    
    def upload (
        self,
        file: Union[str, Path, BytesIO],
        type: UploadType,
        name: str=None,
        data_url_limit: int=None,
        key: str=None,
        verbose: bool=False
    ) -> str:
        """
        Upload a file and return the URL.

        Parameters:
            file (str | Path | BytesIO): Input file.
            type (UploadType): File type.
            name (str): File name. This MUST be provided if `file` is not a file path.
            data_url_limit (int): Return a data URL if the file is smaller than this limit (in bytes).
            key (str): File key. This is useful for grouping related files.
            verbose (bool): Print a progress bar for the upload.

        Returns:
            str: Upload URL.
        """
        file = Path(file) if isinstance(file, str) else file
        if isinstance(file, Path):
            return self.__upload_file(file, type, name=name, key=key, data_url_limit=data_url_limit, verbose=verbose)
        else:
            return self.__upload_buffer(file, type, name=name, key=key, data_url_limit=data_url_limit, verbose=verbose)
        
    def __upload_file (
        self,
        file: Path,
        type: UploadType,
        name: str=None,
        key: str=None,
        data_url_limit: int=None,
        verbose: bool=False
    ) -> str:
        # Check file
        assert file.exists(), f"Cannot upload {file.name} because the file does not exist"
        assert file.is_file(), f"Cannot upload {file.name} becaause it does not point to a file"   
        # Create data URL
        mime = guess_mime(file) or "application/octet-stream"
        if file.stat().st_size < (data_url_limit or 0):
            with open(file, mode="rb") as f:
                buffer = BytesIO(f.read())
            return self.__create_data_url(buffer, mime)
        # Upload
        name = name or file.name
        url = self.create_upload_url(name, type, key=key)
        with open_progress(file, mode="rb", description=name, disable=not verbose) as f:
            put(url, data=f, headers={ "Content-Type": mime }).raise_for_status()
        # Return
        return url
    
    def __upload_buffer (
        self,
        file: BytesIO,
        type: UploadType,
        name: str=None,
        key: str=None,
        data_url_limit: int=None,
        verbose: bool=False
    ) -> str:
        # Check name
        assert name, "You must specify the file `name` if the `file` is not a path"
        # Create data URL
        file.seek(0)
        mime = guess_mime(file) or "application/octet-stream"
        size = file.getbuffer().nbytes
        if size < (data_url_limit or 0):
            return self.__create_data_url(file, mime)
        # Upload
        url = self.create_upload_url(name, type, key=key)
        with wrap_file(file, total=size, description=name, disable=not verbose) as f:
            put(url, data=f, headers={ "Content-Type": mime }).raise_for_status()
        # Return
        return url
    
    def __create_data_url (self, file: BytesIO, mime: str) -> str:
        encoded_data = b64encode(file.getvalue()).decode("ascii")
        url = f"data:{mime};base64,{encoded_data}"
        return url


#
#   Function
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from __future__ import annotations
from pydantic import BaseModel
from typing import Any, List, Optional

from .predictor import PredictorType

class Prediction (BaseModel):
    """
    Prediction.

    Members:
        id (str): Prediction ID.
        tag (str): Predictor tag.
        type (PredictorType): Prediction type.
        created (str): Date created.
        results (list): Prediction results.
        latency (float): Prediction latency in milliseconds.
        error (str): Prediction error. This is `null` if the prediction completed successfully.
        logs (str): Prediction logs.
        implementation (str): Predictor implementation. This is only populated for `EDGE` predictions.
        resources (list): Prediction resources. This is only populated for `EDGE` predictions.
        configuration (str): Prediction configuration token. This is only populated for `EDGE` predictions.
    """
    id: str
    tag: str
    type: PredictorType
    created: str
    results: Optional[List[Any]] = None
    latency: Optional[float] = None
    error: Optional[str] = None
    logs: Optional[str] = None
    implementation: Optional[str] = None
    resources: Optional[List[PredictionResource]] = None
    configuration: Optional[str] = None    

class PredictionResource (BaseModel):
    """
    Prediction resource.

    Members:
        id (str): Resource identifier.
        url (str): Resource URL.
    """
    id: str
    url: str


# 
#   Function
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from pathlib import Path
from typing import Dict, List, Union

from ..graph import GraphClient
from ..types import Acceleration, AccessMode, Predictor, PredictorStatus, PredictorType, UploadType
from .storage import StorageService
from .user import PROFILE_FIELDS

class PredictorService:

    def __init__ (self, client: GraphClient, storage: StorageService) -> None:
        self.client = client
        self.storage = storage

    def retrieve (self, tag: str) -> Predictor:
        """
        Retrieve a predictor.

        Parameters:
            tag (str): Predictor tag.

        Returns:
            Predictor: Predictor.
        """
        # Query
        response = self.client.queryfxn(f"""
            query ($input: PredictorInput!) {{
                predictor (input: $input) {{
                    {PREDICTOR_FIELDS}
                }}
            }}
            """,
            { "input": { "tag": tag } }
        )
        # Create predictor
        predictor = response["predictor"]
        predictor = Predictor(**predictor) if predictor else None
        # Return
        return predictor

    def list (
        self,
        owner: str=None,
        status: PredictorStatus=None,
        offset: int=None,
        count: int=None
    ) -> List[Predictor]:
        """
        List the current user's predictors.

        Parameters:
            owner (str): Predictor owner. This defaults to the current user.
            status (PredictorStatus): Predictor status. This defaults to `ACTIVE`.
            offset (int): Pagination offset.
            count (int): Pagination count.

        Returns:
            list: User predictors.
        """
        # Query
        response = self.client.query(f"""
            query ($user: UserInput, $predictors: UserPredictorsInput) {{
                user (input: $user) {{
                    predictors (input: $predictors) {{
                        {PREDICTOR_FIELDS}
                    }}
                }}
            }}
            """,
            {
                "user": { "username": owner } if owner else None,
                "predictors": { "status": status, "offset": offset, "count": count }
            }
        )
        # Check
        user = response["user"]
        if not user:
            return None
        # Create predictors
        predictors = response["user"]["predictors"]
        predictors = [Predictor(**predictor) for predictor in predictors]
        # Return
        return predictors
    
    def search (
        self,
        query: str,
        offset: int=None,
        count: int=None
    ) -> List[Predictor]:
        """
        Search predictors.

        Parameters:
            q (str): Search query.
            offset (int): Pagination offset.
            count (int): Pagination count.

        Returns:
            list: Relevant predictors.
        """
        # Query
        response = self.client.query(f"""
            query ($input: PredictorsInput!) {{
                predictors (input: $input) {{
                    {PREDICTOR_FIELDS}
                }}
            }}
            """,
            { "input": { "query": query, "offset": offset, "count": count } }
        )
        # Create predictors
        predictors = response["predictors"]
        predictors = [Predictor(**predictor) for predictor in predictors]
        # Return
        return predictors
    
    def create (
        self,
        tag: str,
        notebook: Union[str, Path],
        type: PredictorType=None,
        access: AccessMode=None,
        description: str=None,
        media: Union[str, Path]=None,
        acceleration: Acceleration=None,
        environment: Dict[str, str]=None,
        license: str=None,
        overwrite: bool=None
    ) -> Predictor:
        """
        Create a predictor.

        Parameters:
            tag (str): Predictor tag.
            notebook (str | Path): Predictor notebook path or URL.
            type (PredictorType): Predictor type. This defaults to `CLOUD`.
            access (AccessMode): Predictor access mode. This defaults to `PRIVATE`.
            description (str): Predictor description. This must be under 200 characters long.
            media (str | Path): Predictor media path or URL.
            acceleration (Acceleration): Predictor acceleration. This only applies for cloud predictors and defaults to `CPU`. 
            environment (dict): Predictor environment variables.
            license (str): Predictor license URL.
            overwrite (bool): Overwrite any existing predictor with the same tag. Existing predictor will be deleted.

        Returns:
            Predictor: Created predictor.
        """
        # Prepare
        environment = [{ "name": name, "value": value } for name, value in environment.items()] if environment is not None else []
        notebook = self.storage.upload(notebook, UploadType.Notebook) if isinstance(notebook, Path) else notebook
        media = self.storage.upload(media, UploadType.Media) if isinstance(media, Path) else media
        # Query
        response = self.client.query(f"""
            mutation ($input: CreatePredictorInput!) {{
                createPredictor (input: $input) {{
                    {PREDICTOR_FIELDS}
                }}
            }}
            """,
            {
                "input": {
                    "tag": tag,
                    "type": type,
                    "notebook": notebook,
                    "access": access,
                    "description": description,
                    "media": media,
                    "acceleration": acceleration,
                    "environment": environment,
                    "overwrite": overwrite,
                    "license": license
                }
            }
        )
        # Create predictor
        predictor = response["createPredictor"]
        predictor = Predictor(**predictor) if predictor else None
        # Return
        return predictor
    
    def delete (self, tag: str) -> bool:
        """
        Delete a predictor.

        Parameters:
            tag (str): Predictor tag.

        Returns:
            bool: Whether the predictor was successfully deleted.
        """
        # Query
        response = self.client.query(f"""
            mutation ($input: DeletePredictorInput!) {{
                deletePredictor (input: $input)
            }}
            """,
            { "input": { "tag": tag } }
        )
        # Return
        result = response["deletePredictor"]
        return result
    
    def archive (self, tag: str) -> Predictor:
        """
        Archive an active predictor.

        Parameters:
            tag (str): Predictor tag.

        Returns:
            Predictor: Archived predictor.
        """
        # Query
        response = self.client.query(f"""
            mutation ($input: ArchivePredictorInput!) {{
                archivePredictor (input: $input) {{
                    {PREDICTOR_FIELDS}
                }}
            }}
            """,
            { "input": { "tag": tag } }
        )
        # Create predictor
        predictor = response["archivePredictor"]
        predictor = Predictor(**predictor) if predictor else None
        # Return
        return predictor

    
PREDICTOR_FIELDS = f"""
tag
owner {{
    {PROFILE_FIELDS}
}}
name
type
status
access
created
description
card
media
acceleration
signature {{
    inputs {{
        name
        type
        description
        range
        optional
        enumeration {{
            name
            value
        }}
        default_value: defaultValue {{
            data
            type
            shape
        }}
        schema
    }}
    outputs {{
        name
        type
        description
    }}
}}
license
"""


# 
#   Function
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from rich import print
from typer import Exit, Option
from webbrowser import open as open_browser

from ..version import __version__

def _learn_callback (value: bool):
    if value:
        open_browser("https://docs.fxn.ai")
        raise Exit()

def _version_callback (value: bool):
    if value:
        print(__version__)
        raise Exit()

def cli_options (
    learn: bool = Option(None, "--learn", callback=_learn_callback, help="Learn about Function."),
    version: bool = Option(None, "--version", callback=_version_callback, help="Get the Function CLI version.")
):
    pass


# 
#   Function
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from __future__ import annotations
from pydantic import BaseModel

class EnvironmentVariable (BaseModel):
    """
    Predictor environment variable.

    Members:
        name (str): Variable name.
        value (str): Variable value.
    """
    name: str
    value: str


# 
#   Function
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from __future__ import annotations
from enum import Enum
from pydantic import BaseModel, Field
from typing import List, Optional, Tuple, Union

from .dtype import Dtype
from .profile import Profile
from .value import Value

class Predictor (BaseModel):
    """
    Predictor.

    Members:
        tag (str): Predictor tag.
        owner (Profile): Predictor owner.
        name (str): Predictor name.
        type (PredictorType): Predictor type.
        status (PredictorStatus): Predictor status.
        access (AccessMode): Predictor access.
        created (str): Date created.
        description (str): Predictor description.
        card (str): Predictor card.
        media (str): Predictor media URL.
        acceleration (Acceleration): Predictor acceleration. This only applies to cloud predictors.
        signature (Signature): Predictor signature. This is only populated once predictor has been successfully provisioned.
        license (str): Predictor license URL.
    """
    tag: str
    owner: Profile
    name: str
    type: PredictorType
    status: PredictorStatus
    access: AccessMode
    created: str
    description: Optional[str] = None
    card: Optional[str] = None
    media: Optional[str] = None
    acceleration: Optional[Acceleration] = None
    signature: Optional[Signature] = None
    license: Optional[str] = None   

class Signature (BaseModel):
    """
    Predictor signature.

    Members:
        inputs (list): Input parameters.
        outputs (list): Output parameters.
    """
    inputs: List[Parameter]
    outputs: List[Parameter]

class Parameter (BaseModel):
    """
    Predictor parameter.

    Members:
        name (str): Parameter name. This is only populated for input parameters.
        type (Dtype): Parameter type. This is `None` if the type is unknown or unsupported by Function.
        description (str): Parameter description.
        optional (bool): Parameter is optional.
        range (tuple): Parameter value range for numeric parameters.
        enumeration (list): Parameter value choices for enumeration parameters.
        default_value (Value): Parameter default value.
        value_schema (dict): Parameter JSON schema. This is only populated for `list` and `dict` parameters.
    """
    name: Optional[str] = None
    type: Optional[Dtype] = None
    description: Optional[str] = None
    optional: Optional[bool] = None
    range: Optional[Tuple[float, float]] = None
    enumeration: Optional[List[EnumerationMember]] = None
    default_value: Optional[Value] = None
    value_schema: Optional[dict] = Field(None, alias="schema")

class EnumerationMember (BaseModel):
    """
    Prediction parameter enumeration member.

    Members:
        name (str): Enumeration member name.
        value (str | int): Enumeration member value.
    """
    name: str
    value: Union[str, int]

class Acceleration (str, Enum):
    """
    Predictor acceleration.
    """
    CPU = "CPU"
    A40 = "A40"
    A100 = "A100"

class AccessMode (str, Enum):
    """
    Predictor access mode.
    """
    Public = "PUBLIC"
    Private = "PRIVATE"
    Protected = "PROTECTED"

class PredictorType (str, Enum):
    """
    Predictor type.
    """
    Cloud = "CLOUD"
    Edge = "EDGE"

class PredictorStatus (str, Enum):
    """
    Predictor status.
    """
    Provisioning = "PROVISIONING"
    Active = "ACTIVE"
    Invalid = "INVALID"
    Archived = "ARCHIVED"


# 
#   Function
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from IPython.core.interactiveshell import InteractiveShell
from IPython.core.magic import Magics, magics_class, line_magic
from typing import List

@magics_class
class FunctionMagics (Magics):

    @line_magic
    def fxn (self, line: str):
        COMMANDS = {
            "python": self.__python,
            "image": self.__image,
        }
        args = line.split(" ")
        command = COMMANDS.get(args[0], None)
        if command is not None:
            command(args[1:])
        else:
            raise RuntimeError(f"Unrecognized Function command: {args[0]}")

    def __python (self, args: List[str]):
        version = args[0]
        print(f"Predictor will use Python {version} when running on Function")

    def __image (self, args: List[str]):
        image = args[0]
        print(f"Predictor will use base image {image} when running on Function")

def load_ipython_extension (ipython: InteractiveShell):
    ipython.register_magics(FunctionMagics)


# 
#   Function
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from enum import Enum

class Dtype (str, Enum):
    """
    Value type.
    This follows `numpy` dtypes.
    """
    null = "null"
    int8 = "int8"
    int16 = "int16"
    int32 = "int32"
    int64 = "int64"
    uint8 = "uint8"
    uint16 = "uint16"
    uint32 = "uint32"
    uint64 = "uint64"
    float16 = "float16"
    float32 = "float32"
    float64 = "float64"
    bool = "bool"
    string = "string"
    list = "list"
    dict = "dict"
    image = "image"
    video = "video"
    audio = "audio"
    model = "model"
    binary = "binary"


#
#   Function
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from aiohttp import ClientSession
from dataclasses import asdict, is_dataclass
from filetype import guess_mime
from io import BytesIO
from json import dumps, loads
from numpy import array, float32, frombuffer, int32, ndarray
from pathlib import Path
from PIL import Image
from platform import system
from pydantic import BaseModel
from requests import get
from tempfile import NamedTemporaryFile
from typing import Any, AsyncIterator, Dict, List, Union
from uuid import uuid4
from urllib.request import urlopen

from ..graph import GraphClient
from ..types import Dtype, PredictorType, Prediction, Value, UploadType
from .storage import StorageService

class PredictionService:

    def __init__ (self, client: GraphClient, storage: StorageService) -> None:
        self.client = client
        self.storage = storage

    def create (
        self,
        tag: str,
        raw_outputs: bool=False,
        return_binary_path: bool=True,
        data_url_limit: int=None,
        **inputs: Dict[str, Union[ndarray, str, float, int, bool, List, Dict[str, Any], Path, Image.Image, Value]],
    ) -> Prediction:
        """
        Create a prediction.

        Parameters:
            tag (str): Predictor tag.
            raw_outputs (bool): Skip converting output values into Pythonic types.
            return_binary_path (bool): Write binary values to file and return a `Path` instead of returning `BytesIO` instance.
            data_url_limit (int): Return a data URL if a given output value is smaller than this size in bytes. Only applies to `CLOUD` predictions.
            inputs (dict): Input values. Only applies to `CLOUD` predictions.

        Returns:
            Prediction: Created prediction.
        """
        # Serialize inputs
        key = uuid4().hex
        inputs = [{ "name": name, **dict(self.from_value(value, name, key=key)) } for name, value in inputs.items()]
        # Query
        response = self.client.query(f"""
            mutation ($input: CreatePredictionInput!) {{
                createPrediction (input: $input) {{
                    {PREDICTION_FIELDS}
                }}
            }}""",
            { "input": { "tag": tag, "client": self.__get_client_id(), "inputs": inputs, "dataUrlLimit": data_url_limit } }
        )
        # Parse
        prediction = response["createPrediction"]
        prediction = self.__parse_cloud_prediction(prediction, raw_outputs=raw_outputs, return_binary_path=return_binary_path)
        # Return
        return prediction
    
    async def stream (
        self,
        tag: str,
        raw_outputs: bool=False,
        return_binary_path: bool=True,
        data_url_limit: int=None,
        **inputs: Dict[str, Union[ndarray, str, float, int, bool, List, Dict[str, Any], Path, Image.Image, Value]],
    ) -> AsyncIterator[Prediction]:
        """
        Create a streaming prediction.

        NOTE: This feature is currently experimental.

        Parameters:
            tag (str): Predictor tag.
            raw_outputs (bool): Skip converting output values into Pythonic types.
            return_binary_path (bool): Write binary values to file and return a `Path` instead of returning `BytesIO` instance.
            data_url_limit (int): Return a data URL if a given output value is smaller than this size in bytes. Only applies to `CLOUD` predictions.
            inputs (dict): Input values. Only applies to `CLOUD` predictions.

        Returns:
            Prediction: Created prediction.
        """
        # Serialize inputs
        key = uuid4().hex
        inputs = { name: dict(self.from_value(value, name, key=key)) for name, value in inputs.items() }
        # Request
        url = f"{self.client.api_url}/predict/{tag}?stream=true&rawOutputs=true&dataUrlLimit={data_url_limit}"
        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.client.access_key}",
            "fxn-client": self.__get_client_id()
        }
        async with ClientSession(headers=headers) as session:
            async with session.post(url, data=dumps(inputs)) as response:
                async for chunk in response.content.iter_any():
                    payload = loads(chunk)
                    # Check status
                    if response.status >= 400:
                        raise RuntimeError(payload.get("error"))
                    # Yield
                    prediction = self.__parse_cloud_prediction(payload, raw_outputs=raw_outputs, return_binary_path=return_binary_path)
                    yield prediction

    def to_value (
        self,
        value: Value,
        return_binary_path: bool=True
    ) -> Union[str, float, int, bool, ndarray, list, dict, Image.Image, BytesIO, Path]:
        """
        Convert a Function value to a plain value.

        Parameters:
            return_binary_path (str): Write binary values to file and return a `Path` instead of returning `BytesIO` instance.

        Returns:
            str | float | int | bool | ndarray | list | dict | Image.Image | BytesIO | Path: Value.
        """
        # Null
        if value.type == Dtype.null:
            return None
        # Download
        buffer = self.__download_value_data(value.data)
        # Array
        if value.type in [
            Dtype.int8, Dtype.int16, Dtype.int32, Dtype.int64,
            Dtype.uint8, Dtype.uint16, Dtype.uint32, Dtype.uint64,
            Dtype.float16, Dtype.float32, Dtype.float64, Dtype.bool
        ]:
            assert value.shape is not None, "Array value must have a shape specified"
            array = frombuffer(buffer.getbuffer(), dtype=value.type).reshape(value.shape)
            return array if len(value.shape) > 0 else array.item()
        # String
        if value.type == Dtype.string:
            return buffer.getvalue().decode("utf-8")
        # List
        if value.type == Dtype.list:
            return loads(buffer.getvalue().decode("utf-8"))
        # Dict
        if value.type == Dtype.dict:
            return loads(buffer.getvalue().decode("utf-8"))
        # Image
        if value.type == Dtype.image:
            return Image.open(buffer)
        # Binary
        if return_binary_path:
            with NamedTemporaryFile(mode="wb", delete=False) as f:
                f.write(buffer.getbuffer())
            return Path(f.name)
        # Return
        return buffer

    def from_value (
        self,
        value: Union[str, float, int, bool, ndarray, List[Any], Dict[str, any], Path, Image.Image],
        name: str,
        min_upload_size: int=4096,
        key: str=None
    ) -> Value:
        """
        Create a Function value from a plain value.

        Parameters:
            value (str | float | int | bool | ndarray | list | dict | dataclass | Path | PIL.Image): Input value.
            name (str): Value name.
            min_upload_size (int): Values larger than this size in bytes will be uploaded.

        Returns:
            Value: Function value.
        """
        value = self.__try_ensure_serializable(value)
        # None
        if value is None:
            return Value(data=None, type=Dtype.null)
        # Value
        if isinstance(value, Value):
            return value
        # Array
        if isinstance(value, ndarray):
            buffer = BytesIO(value.tobytes())
            data = self.storage.upload(buffer, UploadType.Value, name=name, data_url_limit=min_upload_size, key=key)
            return Value(data=data, type=value.dtype.name, shape=list(value.shape))
        # String
        if isinstance(value, str):
            buffer = BytesIO(value.encode("utf-8"))
            data = self.storage.upload(buffer, UploadType.Value, name=name, data_url_limit=min_upload_size, key=key)
            return Value(data=data, type=Dtype.string)
        # Float
        if isinstance(value, float):
            value = array(value, dtype=float32)
            return self.from_value(value, name, min_upload_size=min_upload_size, key=key)
        # Boolean
        if isinstance(value, bool):
            value = array(value, dtype=bool)
            return self.from_value(value, name, min_upload_size=min_upload_size, key=key)
        # Integer
        if isinstance(value, int):
            value = array(value, dtype=int32)
            return self.from_value(value, name, min_upload_size=min_upload_size, key=key)
        # List
        if isinstance(value, list):
            value = dumps(value)
            buffer = BytesIO(value.encode("utf-8"))
            data = self.storage.upload(buffer, UploadType.Value, name=name, data_url_limit=min_upload_size, key=key)
            return Value(data=data, type=Dtype.list)
        # Dict
        if isinstance(value, dict):
            value = dumps(value)
            buffer = BytesIO(value.encode("utf-8"))
            data = self.storage.upload(buffer, UploadType.Value, name=name, data_url_limit=min_upload_size, key=key)
            return Value(data=data, type=Dtype.dict)
        # Image
        if isinstance(value, Image.Image):
            buffer = BytesIO()
            format = "PNG" if value.mode == "RGBA" else "JPEG"
            value.save(buffer, format=format)
            data = self.storage.upload(buffer, UploadType.Value, name=name, data_url_limit=min_upload_size, key=key)
            return Value(data=data, type=Dtype.image)
        # Binary
        if isinstance(value, Path):
            assert value.exists(), "Value does not exist at the given path"
            assert value.is_file(), "Value path must point to a file, not a directory"
            value = value.expanduser().resolve()
            data = self.storage.upload(value, UploadType.Value, name=name, data_url_limit=min_upload_size, key=key)
            dtype = self.__get_file_dtype(value)
            return Value(data=data, type=dtype)
        # Unsupported
        raise RuntimeError(f"Cannot create Function value '{name}' for value {value} of type {type(value)}")
    
    def __parse_cloud_prediction (
        self,
        prediction: Dict[str, Any],
        raw_outputs: bool=False,
        return_binary_path: bool=True
    ) -> Prediction:
        # Check null
        if not prediction:
            return None
        # Check type
        if prediction["type"] != PredictorType.Cloud:
            return prediction
        # Gather results
        if "results" in prediction and prediction["results"] is not None:
            prediction["results"] = [Value(**value) for value in prediction["results"]]
            if not raw_outputs:
                prediction["results"] = [self.to_value(value, return_binary_path=return_binary_path) for value in prediction["results"]]
        # Return
        return Prediction(**prediction)

    def __get_file_dtype (self, path: Path) -> Dtype:
        mime = guess_mime(str(path))
        if not mime:
            return Dtype.binary
        if mime.startswith("image"):
            return Dtype.image
        if mime.startswith("video"):
            return Dtype.video
        if mime.startswith("audio"):
            return Dtype.audio
        if path.suffix in [".obj", ".gltf", ".glb", ".fbx", ".usd", ".usdz", ".blend"]:
            return Dtype._3d
        return Dtype.binary

    def __download_value_data (self, url: str) -> BytesIO:
        if url.startswith("data:"):
            with urlopen(url) as response:
                return BytesIO(response.read())
        response = get(url)
        result = BytesIO(response.content)
        return result
    
    @classmethod
    def __try_ensure_serializable (cls, value: Any) -> Any:
        if value is None:
            return value
        if isinstance(value, Value): # passthrough
            return value
        if isinstance(value, list):
            return [cls.__try_ensure_serializable(x) for x in value]
        if is_dataclass(value) and not isinstance(value, type):
            return asdict(value)
        if isinstance(value, BaseModel):
            return value.model_dump(mode="json")
        return value
    
    @classmethod
    def __get_client_id (self) -> str:
        id = system()
        if id == "Darwin":
            return "macos"
        if id == "Linux":
            return "linux"
        if id == "Windows":
            return "windows"
        raise RuntimeError(f"Function cannot make predictions on the {id} platform")
    

PREDICTION_FIELDS = f"""
id
tag
type
created
implementation
configuration
resources {{
    id
    url
}}
results {{
    data
    type
    shape
}}
latency
error
logs
"""


#
#   Function
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from __future__ import annotations
from pydantic import BaseModel
from typing import List, Optional, Union

from .dtype import Dtype

class Value (BaseModel):
    """
    Prediction value.

    Members:
        data (str): Value URL. This can be a web URL or a data URL.
        type (Dtype): Value type.
        shape (list): Value shape. This is `None` if shape information is not available or applicable.
    """
    data: Union[str, None]
    type: Dtype
    shape: Optional[List[int]] = None